# Copyright (c) 2022 Manfred SCHLAEGL <manfred.schlaegl@gmx.at>
#
# SPDX-License-Identifier: BSD 3-clause "New" or "Revised" License
#

from _libfstapi import ffi, lib
from collections import namedtuple


def string(val):
    """Converts ffi cdata to python string"""
    if val == ffi.NULL:
        return ""
    return ffi.string(val).decode("UTF-8")


def get_scopes_signals2(fst):
    """Iterates the hierarchy (using fstReaderIterateHierRewind and
    fstReaderIterateHier) and returns a list containing all scope names
    and a named tuple containing two dictionaries which describe the signals:
    The "by_name" dictionary maps from signal names, the "by_handle" from
    signal handles to a corresponding named tuple containing signal name,
    length and handle.
    Note that signals with different name may have the same handle,
    therefore "by_handle" may have less entries than "by_name".
    """

    scopes = []
    signals_by_name = {}
    signals_by_handle = {}
    cur_scope = ""
    last_scopes = []
    Signal = namedtuple("Signal", "name length handle")
    Signals = namedtuple("Signals", "by_name by_handle")

    lib.fstReaderIterateHierRewind(fst)
    while True:
        fstHier = lib.fstReaderIterateHier(fst)
        if fstHier == ffi.NULL:
            break

        if fstHier.htyp == lib.FST_HT_SCOPE:
            last_scopes.append(cur_scope)

            # add new scope
            if cur_scope != "":
                cur_scope += "."
            cur_scope += string(fstHier.u.scope.name)
            scopes.append(cur_scope)

        elif fstHier.htyp == lib.FST_HT_UPSCOPE:
            # restore last scope
            cur_scope = last_scopes.pop()

        elif fstHier.htyp == lib.FST_HT_VAR:
            # add new variable with handle
            var_name = cur_scope + "." + string(fstHier.u.var.name)
            handle = fstHier.u.var.handle
            signal = Signal(var_name, fstHier.u.var.length, handle)
            signals_by_name[var_name] = signal
            # add first signal name for handle
            signals_by_handle.setdefault(handle, signal)
            signals = Signals(signals_by_name, signals_by_handle)

    return (scopes, signals)


def get_scopes_signals(fst):
    """(Deprecated) Iterate the hierarchy (using fstReaderIterateHierRewind and
    fstReaderIterateHier) and return a list containing all scopes names
    and a dictionary containing all signal names with corresponding handles

    This function is deprecated! Use get_scopes_signals2 instead!
    """
    (scopes, signals) = get_scopes_signals2(fst)
    return (scopes, {key: value.handle for (key, value) in signals.by_name.items()})


def get_signal_name_by_handle(signals, handle):
    """(Deprecated) Returns the first matching signal name from the
    given signals dictionary for the given handle

    This function is deprecated! Use get_scopes_signals2 and the returned
    "by_handle" signal dictionary instead!
    """
    return list(signals.keys())[list(signals.values()).index(handle)]


def fstReaderIterBlocks(
    fst, value_change_callback, user_callback_data=None, vcdhandle=None
):
    """Wrapped version of fstReaderIterBlocks. Allows the use of any
    normal python function as callback (with slight overhead)"""

    if vcdhandle is None:
        vcdhandle = ffi.NULL

    # wrap python callbacks and callback data
    data = ffi.new_handle((value_change_callback, None, user_callback_data))

    # call with wrapper
    return lib.fstReaderIterBlocks(
        fst, lib.pylibfst_wrapped_value_change_callback, data, vcdhandle
    )


def fstReaderIterBlocks2(
    fst,
    value_change_callback,
    value_change_callback_varlen,
    user_callback_data=None,
    vcdhandle=None,
):
    """Wrapped version of fstReaderIterBlocks2. Allows the use of any
    normal python function as callback (with slight overhead)"""

    if vcdhandle is None:
        vcdhandle = ffi.NULL

    # wrap python callbacks and callback data
    data = ffi.new_handle(
        (value_change_callback, value_change_callback_varlen, user_callback_data)
    )

    # call with wrapper
    return lib.fstReaderIterBlocks2(
        fst,
        lib.pylibfst_wrapped_value_change_callback,
        lib.pylibfst_wrapped_value_change_callback_varlen,
        data,
        vcdhandle,
    )


@ffi.def_extern()
def pylibfst_wrapped_value_change_callback(data, time, facidx, value):
    """INTERNAL USE ONLY!
    Callback wrapper for fstReaderIterBlocks and fstReaderIterBlocks2.
    Unwraps python function and data and calls the python function with data"""

    # unwrap python callback and data
    data = ffi.from_handle(data)
    # call wrapped callback
    data[0](data[2], time, facidx, value)


@ffi.def_extern()
def pylibfst_wrapped_value_change_callback_varlen(data, time, facidx, value, length):
    """INTERNAL USE ONLY!
    Callback wrapper for fstReaderIterBlocks2.
    Unwraps python function and data and calls the python function with data"""

    # unwrap python callback and data
    data = ffi.from_handle(data)
    # call wrapped callback
    data[1](data[2], time, facidx, value, length)



# Copyright (c) 2022 Manfred SCHLAEGL <manfred.schlaegl@gmx.at>
#
# SPDX-License-Identifier: BSD 3-clause "New" or "Revised" License
#

import os
from cffi import FFI

ffibuilder = FFI()

with open(os.path.join(os.path.dirname(__file__), "libfstapi.cdef")) as fp:
    cdef = fp.read()
ffibuilder.cdef(cdef)
ffibuilder.set_source(
    "_libfstapi",
    """
    // the C headers of the library
    #include "fstapi.h"
    #include "fstext.h"
""",
    # fstapi(libfstapi.a) is created by setup.py using
    # cmake and the original CMakeLists.txt of fst
    libraries=["fst/fstapi", "z"],
    include_dirs=["fst"],
)

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



# Copyright (c) 2022 Manfred SCHLAEGL <manfred.schlaegl@gmx.at>
#
# SPDX-License-Identifier: BSD 3-clause "New" or "Revised" License
#

import sys
import pylibfst

# cffi style callbacks


@pylibfst.ffi.def_extern()
def pylibfst_value_change_callback(user_callback_data_pointer, time, facidx, value):

    print(
        "value_change_callback "
        + str(time)
        + " "
        + str(facidx)
        + " "
        + signals.by_handle[facidx].name
        + " "
        + pylibfst.helpers.string(value)
    )


@pylibfst.ffi.def_extern()
def pylibfst_value_change_callback_varlen(
    user_callback_data_pointer, time, facidx, value, length
):

    print(
        "value_change_callback_varlen "
        + str(time)
        + " "
        + str(facidx)
        + " "
        + signals.by_handle[facidx].name
        + " "
        + pylibfst.helpers.string(value)
        + " "
        + str(length)
    )


if len(sys.argv) != 2:
    print(
        "IterBlocks_callback (pylibfst example) (C) 2022 Manfred SCHLAEGL <manfred.schlaegl@gmx.at>\n"
    )
    print("Usage: " + sys.argv[0] + " <fstfile>\n")
    print("Example: " + sys.argv[0] + " counter.fst\n")
    sys.exit(1)
filename = sys.argv[1]

fst = pylibfst.lib.fstReaderOpen(filename.encode("UTF-8"))
if fst == pylibfst.ffi.NULL:
    print("Unable to open file '" + filename + "'!")
    sys.exit(1)


(scopes, signals) = pylibfst.get_scopes_signals2(fst)

pylibfst.lib.fstReaderSetFacProcessMaskAll(fst)

print("fstReaderIterBlocks")
ret = pylibfst.lib.fstReaderIterBlocks(
    fst,
    pylibfst.lib.pylibfst_value_change_callback,
    pylibfst.ffi.NULL,
    pylibfst.ffi.NULL,
)
print("ret " + str(ret))

print("fstReaderIterBlocks2")
ret = pylibfst.lib.fstReaderIterBlocks2(
    fst,
    pylibfst.lib.pylibfst_value_change_callback,
    pylibfst.lib.pylibfst_value_change_callback_varlen,
    pylibfst.ffi.NULL,
    pylibfst.ffi.NULL,
)
print("ret " + str(ret))

pylibfst.lib.fstReaderClose(fst)
print("done")



# Copyright (c) 2022 Manfred SCHLAEGL <manfred.schlaegl@gmx.at>
#
# SPDX-License-Identifier: BSD 3-clause "New" or "Revised" License
#

import sys
import pylibfst


def printi(indent, *args):
    for i in range(indent):
        print("  ", end="")
    print("+ ", end="")
    print(*args)


def dumpHierachryEntryScope(indent, fstHier):
    printi(indent, "Scope:")
    indent += 1
    printi(indent, "Type:          " + str(fstHier.u.scope.typ))
    printi(indent, "Name:          " + pylibfst.helpers.string(fstHier.u.scope.name))
    printi(
        indent,
        "Component:     " + str(pylibfst.helpers.string(fstHier.u.scope.component)),
    )
    return indent


def dumpHierachryEntryVar(indent, fstHier):
    printi(indent, "Var:")
    indent += 1
    printi(indent, "Type:          " + str(fstHier.u.var.typ))
    printi(indent, "Name:          " + pylibfst.helpers.string(fstHier.u.var.name))
    printi(indent, "Direction:     " + str(fstHier.u.var.direction))
    printi(indent, "SVT Workspace: " + str(fstHier.u.var.svt_workspace))
    printi(indent, "SDT Workspace: " + str(fstHier.u.var.sdt_workspace))
    printi(indent, "SXT Workspace: " + str(fstHier.u.var.sxt_workspace))
    printi(indent, "Length:        " + str(fstHier.u.var.length))
    printi(indent, "fstHandle:     " + str(fstHier.u.var.handle))
    printi(indent, "is_alias:      " + str(fstHier.u.var.is_alias))
    indent -= 1
    return indent


def dumpHierachryEntryAttrBegin(indent, fstHier):
    printi(indent, "AttrBegin:")
    indent += 1
    printi(indent, "Type:          " + str(fstHier.u.attr.typ))
    printi(indent, "Sub Type:      " + str(fstHier.u.attr.subtyp))
    printi(indent, "Name:          " + pylibfst.helpers.string(fstHier.u.attr.name))
    printi(indent, "Arg:           " + str(fstHier.u.attr.arg))
    printi(indent, "Arg from Name: " + str(fstHier.u.attr.arg_from_name))
    return indent


def dumpHierachryEntryTreeBegin(indent, fstHier):
    indent += 1
    print(" + TreeBegin: Not Implemented")
    return indent


def dumpHierachyEntry(indent, fstHier):

    if fstHier.htyp == pylibfst.lib.FST_HT_SCOPE:
        indent = dumpHierachryEntryScope(indent, fstHier)

    elif fstHier.htyp == pylibfst.lib.FST_HT_UPSCOPE:
        indent -= 1

    elif fstHier.htyp == pylibfst.lib.FST_HT_VAR:
        indent = dumpHierachryEntryVar(indent, fstHier)

    elif fstHier.htyp == pylibfst.lib.FST_HT_ATTRBEGIN:
        indent = dumpHierachryEntryAttrBegin(indent, fstHier)

    elif fstHier.htyp == pylibfst.lib.FST_HT_ATTREND:
        indent -= 1

    elif fstHier.htyp == pylibfst.lib.FST_HT_TREEBEGIN:
        indent = dumpHierachryEntryTreeBegin(indent, fstHier)

    elif fstHier.htyp == pylibfst.lib.FST_HT_TREEEND:
        indent -= 1

    else:
        print("Invalid htyp " + str(fstHier.htyp))

    return indent


def dumpHierachy(fst):

    print("DUMP HIRARCHY: ")
    indent = 0
    pylibfst.lib.fstReaderIterateHierRewind(fst)

    while True:
        fstHier = pylibfst.lib.fstReaderIterateHier(fst)
        if fstHier == pylibfst.ffi.NULL:
            break
        indent = dumpHierachyEntry(indent, fstHier)


def dump(fst):

    verStr = pylibfst.lib.fstReaderGetVersionString(fst)
    print("Version String:           " + pylibfst.helpers.string(verStr))

    date = pylibfst.lib.fstReaderGetDateString(fst)
    print("Date String:              " + pylibfst.helpers.string(date))

    fileType = pylibfst.lib.fstReaderGetFileType(fst)
    print("File Type:                " + str(fileType))

    varCount = pylibfst.lib.fstReaderGetVarCount(fst)
    print("Var Count:                " + str(varCount))

    scopeCount = pylibfst.lib.fstReaderGetScopeCount(fst)
    print("Scope Count:              " + str(scopeCount))

    aliasCount = pylibfst.lib.fstReaderGetAliasCount(fst)
    print("Alias Count:              " + str(aliasCount))

    startTime = pylibfst.lib.fstReaderGetStartTime(fst)
    endTime = pylibfst.lib.fstReaderGetEndTime(fst)
    print("Start Time:               " + str(startTime))
    print("End Time:                 " + str(endTime))

    timeScale = pylibfst.lib.fstReaderGetTimescale(fst)
    print("Time Scale:               " + str(timeScale))

    timeZero = pylibfst.lib.fstReaderGetTimezero(fst)
    print("Time Zero:                " + str(timeZero))

    valChSecCnt = pylibfst.lib.fstReaderGetValueChangeSectionCount(fst)
    print("Value Change Section Cnt: " + str(valChSecCnt))

    dumpHierachy(fst)


def dump_signals(fst, signals):

    # get timestamps of all signal changes
    pylibfst.lib.fstReaderSetFacProcessMaskAll(fst)
    timestamps = pylibfst.lib.fstReaderGetTimestamps(fst)

    for signal in signals.by_name.values():
        print("'" + signal.name + "'; ", end="")
    print()

    buf = pylibfst.ffi.new("char[256]")
    for ts in range(timestamps.nvals):
        time = timestamps.val[ts]
        print("{: >5d}; ".format(time), end="")
        for signal in signals.by_name.values():
            val = pylibfst.helpers.string(
                pylibfst.lib.fstReaderGetValueFromHandleAtTime(
                    fst, time, signal.handle, buf
                )
            )
            print(str(val) + "; ", end="")
        print()

    pylibfst.lib.fstReaderFreeTimestamps(timestamps)


if len(sys.argv) != 2:
    print(
        "dumpfst (pylibfst example) (C) 2022 Manfred SCHLAEGL <manfred.schlaegl@gmx.at>\n"
    )
    print("Usage: " + sys.argv[0] + " <fstfile>\n")
    print("Example: " + sys.argv[0] + " counter.fst\n")
    sys.exit(1)
filename = sys.argv[1]

fst = pylibfst.lib.fstReaderOpen(filename.encode("UTF-8"))
if fst == pylibfst.ffi.NULL:
    print("Unable to open file '" + filename + "'!")
    sys.exit(1)

dump(fst)
print()

(scopes, signals) = pylibfst.get_scopes_signals2(fst)
print("Scopes:  " + str(scopes))
print("Signals:")
for signal in signals.by_name.values():
    print("    " + str(signal))
print()

dump_signals(fst, signals)
print()

pylibfst.lib.fstReaderClose(fst)
print("done")
